#ifndef CALC_H
#define CALC_H

#define NUMBER_TYPE   50
#define POINTER_TYPE  51

#define T_POWER 0
#define T_MULT  1
#define T_DIV   2
#define T_MOD   3
#define T_PLUS  4
#define T_MINUS 5
#define T_LT 6
#define T_LE 7
#define T_GT 8
#define T_GE 9
#define T_EQ 10
#define T_NEQ 11
#define T_AND 12
#define T_OR 13

#define TOP_BINARY 13

#define BR_CLOSE 14
#define BRACKET 15

#define T_SINH 16
#define T_COSH 17
#define T_TANH 18
#define T_SIN  19
#define T_COS  20
#define T_TAN  21
#define T_EXP  22
#define T_SQR  23
#define T_SQRT 24
#define T_THETA 25
#define T_SIGMA  26
#define T_LOG   27
#define T_LOG10 28
#define T_CUBE  29
#define T_ATAN  30
#define T_INT  31
#define T_NOT  32


#define TYPE_NUM 33

const char type_id[TYPE_NUM][7] = { "^", "*", "/", "mod", "+", "-", "<",
				    "<=", ">", ">=", "==", "!=", "and", "or", ")", "(",
                        "sinh(", "cosh(", "tanh(", "sin(", "cos(", "tan(", 
                        "exp(", "sqr(", "sqrt(", "theta(", "sigma(", "log(", "log10(", "cube(", 
                        "atan(", "int(", "not(" };

inline bool mayBeUnary(int tp)   { if ( (tp == T_PLUS) || (tp == T_MINUS) ) return true; else return false; } 
inline bool isBinaryOnly(int tp) { if ( (tp <= TOP_BINARY) && !mayBeUnary(tp) ) return true; else return false; }
inline bool isOperand(int tp)    { if ( (tp == NUMBER_TYPE) || (tp == POINTER_TYPE) ) return true; else return false; }

inline bool isExpressionStart(char *s) {
  if ( equal(s,ASSIGN_LABEL) || equal(s,VAR_LABEL) || equal(s,ODE_LABEL) || equal(s,IF_TOKEN) || 
       equal(s, "print" ) || equal(s,"append") || equal(s,"current") || equal(s,"tortuosity") ) 
  return true; else return false; }

inline bool isExpressionStop(char *s) {
  if (equal(s,CARR_RET_TOKEN) || equal(s,"to") || equal(s,"step") || equal(s, "then" ) ) return true; else return false;}

int get_max_term_num(TokenString &, int);

double *ResolveID(char *, double **, FieldObj *Ca, BufferArray *Bufs, class KineticObj *KO);
char   *ResolvePtr(double *ptr, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO);

bool isFloat(const char *s);
bool isInt(const char *s);

//*************************************************************************************

struct TermStruct
{
  char   type;
  double val;
  double *ptr;
};

//*************************************************************************************

class ExpressionObj
{

 public:

  struct TermStruct *term_array;
  int    term_num;

  ExpressionObj(int n = 0)  { term_array = new TermStruct[term_num = n]; }

  ExpressionObj(const ExpressionObj &EO) { 
    term_array = new TermStruct[term_num = EO.term_num];
    *this = EO;
  }

  void operator=(const ExpressionObj &EO) {
    for (int i = 0; i < term_num; i++) term_array[i] = EO.term_array[i]; }

  ~ExpressionObj()    { delete [] term_array; }
  void   Optimize();
  double Evaluate();
  void print(FILE *f = stderr, FieldObj *Ca=0, BufferArray *Bufs=0, KineticObj *KO=0);

  friend class KinExpression;
};

//*************************************************************************************

class KinExpression : public ExpressionObj
{
 public:

  KinExpression(TokenString &, int, double * = 0, const char * = "t", FieldObj * = 0, BufferArray * = 0,
                class KineticObj * = 0, int * = 0);

  KinExpression() : ExpressionObj(0) { } 
};

//*************************************************************************************

#endif
